package org.artifactory.metadata.service.provider;

import com.google.common.collect.ImmutableSet;
import lombok.Data;
import org.apache.commons.lang.StringUtils;
import org.artifactory.api.module.ModuleInfo;
import org.artifactory.aql.api.domain.sensitive.AqlApiItem;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.md.Properties;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.service.InternalRepositoryService;
import org.jfrog.client.util.PathUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import javax.annotation.Nonnull;
import java.util.HashSet;
import java.util.Optional;
import java.util.Set;

import static org.artifactory.metadata.service.provider.MavenGroupUtils.*;

/**
 * @author Yuval Reches
 */
@Component
public class SbtMetadataProvider extends AbstractMetadataProvider {
    private static final Logger log = LoggerFactory.getLogger(SbtMetadataProvider.class);
    private static final RepoType type = RepoType.SBT;
    private static final String LEAD_FILE_EXTENSION = "xml";
    private static final String LEAD_FILE_EXTENSION_POM = "pom";
    private InternalRepositoryService repositoryService;

    @Autowired
    SbtMetadataProvider(InternalRepositoryService repositoryService) {
        super(type);
        setTypeSpecificMetadataTemplate(new SbtTypeSpecificMetadataTemplate());
        this.repositoryService = repositoryService;
    }

    /**
     * Lead artifact for SBT can be either *.pom file, or ivy.xml file (according to SBT documentation that's the only
     * option for Ivy, while for Ivy itself the file's name can be many variations of ivy{.+}.xml)
     */
    @Override
    public boolean isLeadArtifact(RepoPath repoPath) {
        return isPomLead(repoPath) || isIvyLead(repoPath);
    }

    /**
     * Looking for version file list according to type of lead artifact.
     * See documentation under the {@link AbstractMetadataProvider#buildVersionFilesQuery} of the Ivy or Maven type.
     */
    @Override
    public Optional<AqlApiItem> buildVersionFilesQuery(@Nonnull RepoPath repoPath) {
        if (isPomLead(repoPath)) {
            return getPomVersionFileQuery(repoPath, repositoryService);
        } else if (isIvyLead(repoPath)) {
            return getIvyVersionFileQuery(repoPath);
        }
        log.debug("The lead artifact '{}' is not of type pom or ivy.xml, can't generate a file list", repoPath);
        return Optional.empty();
    }

    @Override
    protected String getType() {
        return type.getType();
    }

    private boolean isIvyLead(RepoPath repoPath) {
        return repoPath.getName().equals("ivy.xml");
    }

    private boolean isPomLead(RepoPath repoPath) {
        String extension = PathUtils.getExtension(repoPath.getPath());
        return StringUtils.isNotBlank(extension) && extension.equals(LEAD_FILE_EXTENSION_POM);
    }

    @Data
    class SbtTypeSpecificMetadataTemplate implements TypeSpecificMetadataTemplate {
        @Override
        public Set<String> getExcludedUserProperties() {
            return new HashSet<>();
        }

        @Override
        public Set<String> getLeadFileExtensions() {
            return ImmutableSet.of(LEAD_FILE_EXTENSION, LEAD_FILE_EXTENSION_POM);
        }

        @Override
        public String getName(Properties properties, RepoPath repoPath) {
            return getModuleInfo(repoPath, repositoryService).map(ModuleInfo::getModule).orElse(null);
        }

        @Override
        public String getVersion(RepoPath repoPath, Properties properties) {
            return getModuleInfo(repoPath, repositoryService).map(ModuleInfo::getBaseRevision).orElse(null);
        }

        @Override
        public String getPkgidSuffix(RepoPath repoPath, Properties properties) {
            return getModuleInfo(repoPath, repositoryService).map(info -> getPkgIdSuffixFromLayout(info, repoPath))
                    .orElse(null);
        }
    }
}
