/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.util;

import lombok.NonNull;
import lombok.experimental.UtilityClass;
import org.artifactory.common.home.ArtifactoryHome;
import org.jfrog.common.TomcatUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * A helper class that attempts to detect the http port Artifactory is listening on.
 *
 * @author Yossi Shaul
 */
@UtilityClass
public class ListeningPortDetector {
    private static final Logger log = LoggerFactory.getLogger(ListeningPortDetector.class);

    public static final String SYS_ARTIFACTORY_PORT = "artifactory.port";
    public static final String SYS_MAINTENANCE_ARTIFACTORY_PORT = "artifactory.tomcat.maintenanceConnector.port";
    private static final int ARTIFACTORY_PREFERRED_PORT = 8081;
    private static final int ARTIFACTORY_MAINTENANCE_PREFERRED_PORT = 8091;

    /**
     * @return The detected http port or -1 if not found
     */
    public static int detect() {
        int port = detectFromSystemProperty(SYS_ARTIFACTORY_PORT);
        if (port == -1) {
            try {
                port = ArtifactoryHome.get().getSysConfig().getInt("artifactory.entrypoints.internalPort", -1);
            } catch (IllegalStateException e) {
                log.debug("Unable to get Artifactory home.", e);
            }
        }
        if (port == -1) {
            port = detectFromTomcatMBean();
        }
        return port;
    }

    /**
     * @return Configured maintenance port, or ARTIFACTORY_MAINTENANCE_PREFERRED_PORT as default
     *
     * @throws IllegalStateException if ArtifactoryHome is not initialized
     */
    public static int getMaintenancePort() {
        return ArtifactoryHome.get().getSysConfig()
                .getInt(SYS_MAINTENANCE_ARTIFACTORY_PORT, ARTIFACTORY_MAINTENANCE_PREFERRED_PORT);
    }

    private static int detectFromSystemProperty(@NonNull String propertyName) {
        String portFromSystem = System.getProperty(propertyName);
        if (portFromSystem != null) {
            try {
                return Integer.parseUnsignedInt(portFromSystem);
            } catch (NumberFormatException e) {
                log.warn("Unable to parse Artifactory port from system property: {}={}", propertyName, portFromSystem);
            }
        }
        return -1;
    }

    private static int detectFromTomcatMBean() {
        // the detector will return the port matches the preferred port or the first HTTP/1.1 connector
        int port = -1;
        try {
            port = TomcatUtils
                    .getMainConnector("http", ARTIFACTORY_PREFERRED_PORT, ARTIFACTORY_MAINTENANCE_PREFERRED_PORT)
                    .getPort();
        } catch (IllegalStateException e) {
            log.error("Could not detect port from Tomcat MBean: {}", e.getMessage());
            log.debug(e.getMessage(), e);
        }
        return port;
    }
}
