package org.artifactory.storage.db;

import com.google.common.collect.Lists;
import org.artifactory.common.config.db.ArtifactoryDbProperties;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.converter.ArtifactoryConverterAdapter;
import org.artifactory.storage.db.version.ArtifactoryDBVersion;
import org.artifactory.storage.db.version.converter.DBConverter;
import org.artifactory.storage.db.version.converter.DbDetails;
import org.artifactory.storage.db.version.converter.DbSqlConverterUtil;
import org.artifactory.version.ArtifactoryVersion;
import org.artifactory.version.CompoundVersionDetails;
import org.jfrog.storage.DbType;
import org.jfrog.storage.JdbcHelper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nullable;
import java.util.Arrays;
import java.util.List;
/**
 *
 * @author Omri Ziv
 */
public class DatabaseConverterAdapter implements ArtifactoryConverterAdapter {
    private static final Logger log = LoggerFactory.getLogger(DatabaseConverterAdapter.class);
    private ArtifactoryHome artifactoryHome;
    private DbDetails dbDetails;

    public DatabaseConverterAdapter(ArtifactoryHome artifactoryHome,
            DbDetails dbDetails) {
        this.artifactoryHome = artifactoryHome;
        this.dbDetails = dbDetails;
    }

    @Override
    public void convert(CompoundVersionDetails source, CompoundVersionDetails target) {
        ArtifactoryDbProperties dbProperties = artifactoryHome.getDBProperties();
        JdbcHelper jdbcHelper = dbDetails.getJdbcHelper();

        // All converters of versions above me needs to be executed in sequence
        List<DBConverter> converters = Lists.newArrayList();
        for (ArtifactoryDBVersion dbVersion : ArtifactoryDBVersion.values()) {
            if (dbVersion.getVersion().after(source.getVersion())) {
                converters.addAll(Arrays.asList(dbVersion.getConverters()));
            }
        }

        if (converters.isEmpty()) {
            log.debug("No database converters found between version {} and {}", source, ArtifactoryVersion.getCurrent());
        } else {
            log.info("Starting database conversion from {}({}) to {}({})", source.getVersion().getVersion(), source.getRevision(), ArtifactoryVersion.getCurrent().getVersion(), ArtifactoryVersion.getCurrent().getRevision());
            DbType dbType = dbProperties.getDbType();
            for (DBConverter converter : converters) {
                converter.convert(jdbcHelper, dbType);
            }
            log.info("Finished database conversion from {}({}) to {}({})", source.getVersion().getVersion(), source.getRevision(), ArtifactoryVersion.getCurrent().getVersion(), ArtifactoryVersion.getCurrent().getRevision());
        }

    }

    @Override
    public boolean isInterested(@Nullable CompoundVersionDetails source, CompoundVersionDetails target) {
        return source != null && !source.isCurrent();
    }

    @Override
    public void revert() {

    }

    @Override
    public void backup() {

    }

    @Override
    public void clean() {

    }

    @Override
    public void assertConversionPrecondition(ArtifactoryHome home, CompoundVersionDetails fromVersion,
            CompoundVersionDetails toVersion) {
        DbSqlConverterUtil.verifyDbDetails(dbDetails, artifactoryHome);
    }

}
