package org.artifactory.storage.db.init.migration.itest;

import org.artifactory.storage.db.init.migration.dao.MigrationStatusDao;
import org.artifactory.storage.db.init.migration.entity.DbMigrationStatus;
import org.artifactory.storage.db.itest.DbBaseTest;
import org.fest.util.Collections;
import org.jfrog.common.ClockUtils;
import org.jfrog.common.StreamSupportUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.testng.annotations.Test;

import java.sql.SQLException;
import java.util.Collection;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

import static org.fest.assertions.Assertions.assertThat;
import static org.testng.Assert.*;

/**
 * @author Uriah Levy
 * @author Omri Ziv
 */
@Test
public class MigrationStatusDaoTest extends DbBaseTest {
    @Autowired
    private MigrationStatusDao dao;

    public void testInsertMigration() throws SQLException {
        final long started = ClockUtils.epochMillis();
        dao.insertMigration(new DbMigrationStatus("metadata-service-migration", started, 0, null));
        Optional<DbMigrationStatus> migration = dao.findMigrationById("metadata-service-migration");
        assertTrue(migration.isPresent());
        assertEquals(migration.get().getIdentifier(), "metadata-service-migration");
        assertEquals(migration.get().getStarted(), started);
        assertEquals(migration.get().getFinished(), 0);
    }

    public void testInsertMigrationWithInfoBlob() throws SQLException {
        final long started = ClockUtils.epochMillis();
        dao.insertMigration(new DbMigrationStatus("metadata-service-migration-2", started, 0,
                "{\"migrationTargetNodeId\":, \"123456789\", \"currentEventLogTimestamp\": \"123456\"}".getBytes()));
        Optional<DbMigrationStatus> migration = dao.findMigrationById("metadata-service-migration-2");
        assertTrue(migration.isPresent());
        assertEquals(migration.get().getIdentifier(), "metadata-service-migration-2");
        assertEquals(migration.get().getStarted(), started);
        assertEquals(migration.get().getFinished(), 0);
        assertNotNull(migration.get().getMigrationInfoBlob());
        assertEquals(new String(migration.get().getMigrationInfoBlob()),
                "{\"migrationTargetNodeId\":, \"123456789\", \"currentEventLogTimestamp\": \"123456\"}");
    }

    @Test(dependsOnMethods = "testInsertMigration", expectedExceptions = SQLException.class)
    public void testUniqueConstraint() throws SQLException {
        dao.insertMigration(new DbMigrationStatus("metadata-service-migration", 0, 0, null));
    }

    @Test(dependsOnMethods = "testInsertMigration")
    public void testUpdateMigration() throws SQLException {
        long finished = ClockUtils.epochMillis() + 100;
        dao.updateFinishTimeById("metadata-service-migration", finished);
        Optional<DbMigrationStatus> migration = dao.findMigrationById("metadata-service-migration");
        assertTrue(migration.isPresent());
        assertEquals(migration.get().getIdentifier(), "metadata-service-migration");
        assertEquals(migration.get().getFinished(), finished);
    }

    @Test
    public void testDelete() throws SQLException {
        final long started = ClockUtils.epochMillis();
        dao.insertMigration(new DbMigrationStatus("marker-migration-to-delete", started, 0, null));
        int deleted = dao.delete("marker-migration-to-delete");
        assertEquals(1, deleted);
    }

    @Test
    public void testNotExistedDelete() throws SQLException {
        final long started = ClockUtils.epochMillis();
        try {
            dao.insertMigration(new DbMigrationStatus("marker-migration-to-delete", started, 0, null));
            int deleted = dao.delete("not-existed");
            assertEquals(0, deleted);
        }
        finally {
            // Clean
            int deleted = dao.delete("marker-migration-to-delete");
            assertEquals(1, deleted);
        }
    }

    @Test
    public void testFindAllUnfinishedIn() throws SQLException {
        final long started = ClockUtils.epochMillis();
        try {
            dao.insertMigration(new DbMigrationStatus("marker-migration-1", started, 0, null));
            dao.insertMigration(new DbMigrationStatus("marker-migration-2", started, started + 6, null));
            dao.insertMigration(new DbMigrationStatus("marker-migration-3", started, 0, null));
            dao.insertMigration(new DbMigrationStatus("marker-migration-4", started, 0, null));
            Collection<DbMigrationStatus> dbMigrationStatuses =
                    dao.findAllUnfinishedIn(Collections.list("marker-migration-1", "marker-migration-2",
                            "marker-migration-3"));
            assertEquals(dbMigrationStatuses.size(), 2);
            List<String> names = StreamSupportUtils.stream(dbMigrationStatuses)
                    .map(DbMigrationStatus::getIdentifier)
                    .collect(Collectors.toList());
            assertThat(names).containsExactly("marker-migration-1", "marker-migration-3");
        }
        finally {
            // Clean
            dao.delete("marker-migration-1");
            dao.delete("marker-migration-2");
            dao.delete("marker-migration-3");
            dao.delete("marker-migration-4");
        }
    }
}
