/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.storage.db.locks.service;

import com.google.common.collect.Sets;
import org.apache.commons.lang.StringUtils;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.common.config.db.ArtifactoryDbProperties;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.storage.db.dao.locks.DbDistributeLocksDao;
import org.artifactory.storage.db.dao.locks.LockInfo;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import static org.mockito.Mockito.any;
import static org.mockito.Mockito.anyString;
import static org.mockito.Mockito.eq;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.reset;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

/**
 * @author Shay Bagants
 * @author Mor Merhav
 */
public class DbLocksServiceImplTest extends ArtifactoryHomeBoundTest {

    private DbDistributeLocksDao lockDaoMock;
    private ConcurrentHashMap<String, LockInfo> lockCache;
    private DbLocksService dbLocksService;

    @BeforeMethod
    public void setUp() throws Exception {
        ArtifactoryContext context = mock(ArtifactoryContext.class);
        lockDaoMock = mock(DbDistributeLocksDao.class);
        lockCache = mock(ConcurrentHashMap.class);
        when(context.getServerId()).thenReturn("server");
        ArtifactoryContextThreadBinder.bind(context);
        dbLocksService = new DbLocksServiceImpl(lockDaoMock, new ArtifactoryDbProperties(ArtifactoryHome.get()));
        setMockCacheLockingMap();
    }

    @AfterMethod
    public void tearDown() {
        ArtifactoryContextThreadBinder.unbind();
        reset(lockCache, lockDaoMock);
    }

    @Test
    public void testLockingMapSize() throws SQLException {
        when(lockDaoMock.lockingMapSize(anyString())).thenReturn(3);
        int myCategory = dbLocksService.lockingMapSize("myCategory");
        Assert.assertEquals(myCategory, 3);
        verify(lockDaoMock).lockingMapSize(any());
    }

    @Test(expectedExceptions = IllegalStateException.class, description = "Usually such tests should be avoided, but " +
            "in this case its quite important that we don't swallow SQL exception in cases we don't need to")
    public void testSqlExceptionNotSwallowed() throws SQLException, TimeoutException {
        Connection connection = mock(Connection.class);
        when(lockDaoMock.tryToAcquireLock(any())).thenThrow(new SQLException("bam!"));
        DatabaseMetaData metadata = mock(DatabaseMetaData.class);
        when(metadata.getConnection()).thenReturn(connection);
        ResultSet resultSet = mock(ResultSet.class);
        when(resultSet.next()).thenReturn(true);
        when(metadata.getTables(any(), any(), any(), any())).thenReturn(resultSet);
        when(connection.getMetaData()).thenReturn(metadata);
        when(connection.getSchema()).thenReturn("foo");
        when(lockDaoMock.getLocksConnection()).thenReturn(connection);

        dbLocksService.acquireLock("category", "key", "owner", 1L, TimeUnit.SECONDS);
    }

    @Test(expectedExceptions = RuntimeException.class, expectedExceptionsMessageRegExp = "Failed to query out number " +
            "of locks for.*\\. my reason.*")
    public void testLockingMapSizeErrorMessage() throws SQLException {
        when(lockDaoMock.lockingMapSize("myCategory")).thenThrow(new SQLException("my reason"));
        dbLocksService.lockingMapSize("myCategory");
    }

    @Test(expectedExceptions = IllegalArgumentException.class, expectedExceptionsMessageRegExp = "Servers excluded " +
            "from lock release must not be empty.")
    public void testReleaseOrphanLocksByCategoryExcludingServersEmptyExcludedServerIds() {
        dbLocksService.releaseOrphanLocksByCategoryExcludingServers(Sets.newHashSet(), "category");
    }

    @Test(expectedExceptions = IllegalArgumentException.class, expectedExceptionsMessageRegExp = "Servers excluded " +
            "from lock release must not be empty.")
    public void testReleaseOrphanLocksByCategoryExcludingServersNullExcludedServerIds() {
        dbLocksService.releaseOrphanLocksByCategoryExcludingServers(null, "category");
    }

    @Test(expectedExceptions = IllegalArgumentException.class, expectedExceptionsMessageRegExp = "Lock category " +
            "cannot be empty")
    public void testReleaseOrphanLocksByCategoryExcludingServersEmptyCategory() {
        dbLocksService.releaseOrphanLocksByCategoryExcludingServers(Sets.newHashSet("owner"), StringUtils.EMPTY);
    }

    @Test(expectedExceptions = IllegalArgumentException.class, expectedExceptionsMessageRegExp = "Lock category " +
            "cannot be empty")
    public void testReleaseOrphanLocksByCategoryExcludingServersNullCategory() {
        dbLocksService.releaseOrphanLocksByCategoryExcludingServers(Sets.newHashSet("owner"), null);
    }

    @Test
    public void testReleaseOrphanLocksByCategoryExcludingServers() throws SQLException {
        Set<String> serverIds = Sets.newHashSet("owner");
        LockInfo lockInfo = new LockInfo("category", "key", "owner", 1L, "thread", 111);
        Set<LockInfo> lockInfoSet = Sets.newHashSet(lockInfo);
        when(lockDaoMock.getAllLocksByCategoryNotOwnedBy(serverIds, "category")).thenReturn(lockInfoSet);
        when(lockDaoMock.deleteLockByCategoryAndKey("category", "key")).thenReturn(true);

        dbLocksService.releaseOrphanLocksByCategoryExcludingServers(serverIds, "category");

        verify(lockDaoMock).deleteLockByCategoryAndKey("category", "key");
    }

    @Test(expectedExceptions = IllegalArgumentException.class, expectedExceptionsMessageRegExp = "Servers excluded " +
            "from lock release must not be empty.")
    public void testReleaseOrphanLocksExcludingServersAndCategoriesEmptyExcludedServerIds() {
        dbLocksService.releaseOrphanLocksExcludingServersAndCategories(Sets.newHashSet(), Sets.newHashSet("category"));
    }

    @Test(expectedExceptions = IllegalArgumentException.class, expectedExceptionsMessageRegExp = "Servers excluded " +
            "from lock release must not be empty.")
    public void testReleaseOrphanLocksExcludingServersAndCategoriesNullExcludedServerIds() {
        dbLocksService.releaseOrphanLocksExcludingServersAndCategories(null, Sets.newHashSet("category"));
    }

    @Test(expectedExceptions = IllegalArgumentException.class, expectedExceptionsMessageRegExp = "Categories excluded" +
            " from lock release must not be empty")
    public void testReleaseOrphanLocksExcludingServersAndCategoriesEmptyCategorySet() {
        dbLocksService.releaseOrphanLocksExcludingServersAndCategories(Sets.newHashSet("owner"), Sets.newHashSet());
    }

    @Test(expectedExceptions = IllegalArgumentException.class, expectedExceptionsMessageRegExp = "Categories excluded" +
            " from lock release must not be empty")
    public void testReleaseOrphanLocksExcludingServersAndCategoriesNullCategory() {
        dbLocksService.releaseOrphanLocksExcludingServersAndCategories(Sets.newHashSet("owner"), null);
    }

    @Test
    public void testReleaseOrphanLocksExcludingServersAndCategories() throws SQLException {
        Set<String> serverIds = Sets.newHashSet("owner1");
        Set<String> categories = Sets.newHashSet("category1");
        LockInfo lockInfo = new LockInfo("category2", "key", "owner2", 1L, "thread", 111);
        Set<LockInfo> lockInfoSet = Sets.newHashSet(lockInfo);
        when(lockDaoMock.getAllLocksNotOwnedByAndNotContainedInCategories(serverIds, categories))
                .thenReturn(lockInfoSet);
        when(lockDaoMock.deleteLockByCategoryAndKey("category2", "key")).thenReturn(true);

        dbLocksService.releaseOrphanLocksExcludingServersAndCategories(serverIds, categories);

        verify(lockDaoMock).deleteLockByCategoryAndKey("category2", "key");
    }

    @Test
    public void testReleaseAllLocksByOwnerExcludingKeys() throws SQLException {
        Set<String> excludedKeys = Sets.newHashSet("excludedKey");
        String owner = "owner";

        LockInfo lockInfo1 = new LockInfo("category", "key", owner, 1L, "thread", 111);
        LockInfo lockInfo2 = new LockInfo("category", excludedKeys.iterator().next(), owner, 1L, "thread", 111);
        Set<LockInfo> lockInfoSet = Sets.newHashSet(lockInfo1, lockInfo2);
        when(lockDaoMock.getAllLocksByOwner(eq(owner))).thenReturn(lockInfoSet);
        when(lockDaoMock.deleteLockByCategoryAndKey("category", "key")).thenReturn(true);

        dbLocksService.releaseAllLocksByOwnerExcludingKeys(owner, excludedKeys);

        verify(lockDaoMock).deleteLockByCategoryAndKey("category", "key");
    }

    private void setMockCacheLockingMap() throws NoSuchFieldException, IllegalAccessException {
        Field field = dbLocksService.getClass().getDeclaredField("localLocks");
        field.setAccessible(true);
        Field modifiersField = Field.class.getDeclaredField("modifiers");
        modifiersField.setAccessible(true);
        modifiersField.setInt(field, field.getModifiers() & ~Modifier.FINAL);
        field.set(dbLocksService, lockCache);
    }
}