/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


package org.artifactory.rest.common;

import com.fasterxml.jackson.core.type.TypeReference;
import com.google.common.collect.ImmutableList;
import com.google.common.io.Files;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.FooterMessage;
import org.artifactory.addon.FooterMessage.FooterMessageVisibility;
import org.artifactory.addon.xray.XrayAddon;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.api.storage.StorageQuotaInfo;
import org.artifactory.common.ConstantValues;
import org.artifactory.storage.db.version.v213.V213ConversionFailFunction;
import org.artifactory.security.access.AccessService;
import org.artifactory.storage.StorageService;
import org.artifactory.storage.fs.service.ConfigsService;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.glassfish.jersey.internal.util.collection.StringKeyIgnoreCaseMultivaluedMap;
import org.glassfish.jersey.server.ContainerResponse;
import org.jfrog.common.JsonUtils;
import org.json.JSONException;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.skyscreamer.jsonassert.JSONAssert;
import org.testng.annotations.*;

import javax.ws.rs.core.MultivaluedMap;
import javax.ws.rs.core.Response;
import java.io.File;
import java.io.IOException;
import java.util.List;

import static org.artifactory.addon.FooterLink.URL_PLACEHOLDER;
import static org.artifactory.addon.FooterMessage.createError;
import static org.artifactory.addon.FooterMessage.createWarning;
import static org.fest.assertions.Assertions.assertThat;
import static org.mockito.Mockito.*;
import static org.testng.Assert.*;


/**
 * @author Dan Feldman
 */

@Test
public class GlobalMessageProviderServiceTest extends ArtifactoryHomeBoundTest {
    private static final FooterMessage WARN = createWarning("I'm a teapot!", List.of(), FooterMessageVisibility.admin);
    private static final FooterMessage ERR = createError(
            "I am putting myself to the fullest possible use, which is all I think that any conscious entity can ever hope to do.",
            List.of(), FooterMessageVisibility.admin);

    @Mock
    private AddonsManager addonsManager;
    @Mock
    private StorageService storageService;
    @Mock
    private AuthorizationService authService;
    @Mock
    private XrayAddon xrayAddon;
    @Mock
    private ContainerResponse containerRes;
    @Mock
    private Response res;
    @Mock
    private ConfigsService configsService;
    @Mock
    private AccessService accessService;

    private GlobalMessageProviderServiceImpl msgProvider;
    private MultivaluedMap<String, Object> responseMetadata;
    private File tempFile;

    @BeforeClass
    public void setupMocks() throws IOException {
        MockitoAnnotations.initMocks(this);
        responseMetadata = new StringKeyIgnoreCaseMultivaluedMap<>();
        when(res.getMetadata()).thenReturn(responseMetadata);
        when(containerRes.getHeaders()).thenReturn(responseMetadata);
        when(accessService.isAdminUsingOldDefaultPassword()).thenReturn(false);
        tempFile = File.createTempFile("msg", ".tmp");
        // Setting the cache to be longer than usual so Jenkins slow slaves won't cry :(
        System.setProperty(ConstantValues.globalMessageCacheAgeSecs.getPropertyName(), String.valueOf(30));
    }

    //Current default messages use the addon manager and storage service, apart from their own specific tests they should
    //not return their respective messages.
    @BeforeMethod
    public void setupDefaultMockResponses() {
        //Need a new message provider each time to clear its cache.
        msgProvider = new GlobalMessageProviderServiceImpl();
        msgProvider.setAddonsManager(addonsManager);
        msgProvider.setAccessService(accessService);
        msgProvider.setAuthenticationService(authService);
        msgProvider.setConfigsService(configsService);
        msgProvider.setStorageService(storageService);
        when(addonsManager.getLicenseFooterMessage()).thenReturn(null);
        when(addonsManager.getRemoteRepUnsupportedUrls()).thenReturn(null);
        when(storageService.getStorageQuotaInfo(0)).thenReturn(null);
        when(addonsManager.addonByType(XrayAddon.class)).thenReturn(xrayAddon);
        when(xrayAddon.isXrayEnabled()).thenReturn(false);
        when(authService.isAdmin()).thenReturn(false);
        when(authService.isAnonymous()).thenReturn(false);
        when(configsService.hasConfig(V213ConversionFailFunction.PSQL_NODE_PROPS_INDEX_MISSING_MARKER))
                .thenReturn(false);
    }

    @AfterMethod
    public void clean() {
        responseMetadata.clear();
        homeStub.setProperty(ConstantValues.uiCustomFooterMessageFile, null);
        reset(addonsManager, storageService, authService, xrayAddon);
    }

    public void noMessagesByDefault() {
        assertNoMessages();
    }

    //Anonymous should not see admin messages, this test sets all admin messages (xray, pypi and quota)
    public void testNoAdminMessagesShownToAnon() {
        reset(authService);
        when(authService.isAdmin()).thenReturn(false);
        when(authService.isAnonymous()).thenReturn(true);

        StorageQuotaInfo quota = mock(StorageQuotaInfo.class);
        when(quota.isLimitReached()).thenReturn(true);
        when(storageService.getStorageQuotaInfo(0)).thenReturn(quota);
        when(addonsManager.getRemoteRepUnsupportedUrls()).thenReturn(ERR);
        when(xrayAddon.isXrayEnabled()).thenReturn(true);
        when(xrayAddon.isXrayAliveByHeartbeat()).thenReturn(false);

        assertNoMessages();
    }


    @Test(dataProvider = "providePsqlMessageValues")
    public void testDecorateWithPsqlIndexMismatch(boolean isAdmin, boolean configExists, boolean messageExpected) {
        reset(authService);
        when(authService.isAdmin()).thenReturn(isAdmin);
        when(authService.isAnonymous()).thenReturn(!isAdmin);
        when(configsService.hasConfig(V213ConversionFailFunction.PSQL_NODE_PROPS_INDEX_MISSING_MARKER))
                .thenReturn(configExists);
        if (messageExpected) {
            String globalMessages = msgProvider.getGlobalMessages();
            List<FooterMessage> footerMessages = JsonUtils.getInstance()
                    .readValueRef(globalMessages, new TypeReference<>() {});
            assertNotNull(footerMessages);
            assertThat(footerMessages.get(0).getMessage()).isEqualTo(
                    "An incompatible index has been found for the Artifactory ‘node_props’ database table. This may " +
                            "be caused by certain property values that exceed the allowed length limit, causing " +
                            "creation of the database index to fail during a version upgrade. " +
                            "For instructions on how to fix this error " + URL_PLACEHOLDER);
            assertThat(footerMessages.get(0).getUrls().size()).isEqualTo(1);
            assertThat(footerMessages.get(0).getUrls().get(0).toLink()).isEqualTo(
                    "<a href='https://www.jfrog.com/confluence/display/RTF/Troubleshooting#Troubleshooting-RecoveringfromError:AnincompatibleindexhasbeenfoundfortheArtifactory%E2%80%98node_props%E2%80%99databasetable'>click here</a>");
        } else {
            assertNoMessages();
        }
    }

    @Test
    public void needToChangeAccessAdminPassword() {
        reset(authService);
        when(authService.isAdmin()).thenReturn(true);
        when(authService.isAnonymous()).thenReturn(false);
        when(configsService.hasConfig(V213ConversionFailFunction.PSQL_NODE_PROPS_INDEX_MISSING_MARKER))
                .thenReturn(true);
        when(accessService.isAdminUsingOldDefaultPassword()).thenReturn(true);
        assertThat(msgProvider.getGlobalMessages()).contains(
                "Your Artifactory server is configured with the default admin credentials. " +
                        "It is highly recommended to change the default admin password");
        reset(accessService);
    }

    @DataProvider
    private static Object[][] providePsqlMessageValues() {
        return new Object[][]{
                // isAdminUser, configExists, messageExcepted
                {true, true, true},
                {false, true, false},
                {true, false, false},
                {false, false, false},
        };
    }

    public void testLicenseMsg() {
        reset(authService);
        when(authService.isAdmin()).thenReturn(true);
        when(addonsManager.getLicenseFooterMessage()).thenReturn(WARN);

        assertThat(msgProvider.getGlobalMessages()).isEqualTo(toJson(WARN));
    }

    public void testPypiMsg() {
        reset(authService);
        when(authService.isAdmin()).thenReturn(true);
        when(addonsManager.getRemoteRepUnsupportedUrls()).thenReturn(ERR);

        assertThat(msgProvider.getGlobalMessages()).isEqualTo(toJson(ERR));
    }

    public void testQuotaErr() {
        reset(authService);
        when(authService.isAdmin()).thenReturn(true);
        StorageQuotaInfo quota = mock(StorageQuotaInfo.class);
        when(quota.isLimitReached()).thenReturn(true);
        when(quota.getErrorMessage()).thenReturn("Quota limit reached!");
        when(storageService.getStorageQuotaInfo(0)).thenReturn(quota);

        assertThat(msgProvider.getGlobalMessages()).contains("Quota limit reached!");
    }

    public void testQuotaWarn() {
        reset(authService);
        when(authService.isAdmin()).thenReturn(true);
        StorageQuotaInfo quota = mock(StorageQuotaInfo.class);
        when(quota.isLimitReached()).thenReturn(false);
        when(quota.isWarningLimitReached()).thenReturn(true);
        when(quota.getWarningMessage()).thenReturn("Quota warn!");
        when(storageService.getStorageQuotaInfo(0)).thenReturn(quota);

        assertThat(msgProvider.getGlobalMessages()).contains("Quota warn!");
    }

    public void testXrayLiveWarn() {
        reset(authService);
        when(authService.isAdmin()).thenReturn(true);
        when(xrayAddon.isXrayEnabled()).thenReturn(true);
        when(xrayAddon.isXrayAliveByHeartbeat()).thenReturn(false);

        assertThat(msgProvider.getGlobalMessages()).contains("JFrog Xray is unavailable");
    }

    public void testXrayVersionWarn() {
        reset(authService);
        when(authService.isAdmin()).thenReturn(true);
        when(xrayAddon.isXrayEnabled()).thenReturn(true);
        when(xrayAddon.isXrayAliveByHeartbeat()).thenReturn(true);
        when(xrayAddon.isXrayVersionValidByHeartbeat()).thenReturn(false);

        assertThat(msgProvider.getGlobalMessages()).contains("is incompatible with this");
    }

    @Test
    public void testInvalidateCache() {
        reset(authService);
        when(authService.isAdmin()).thenReturn(true);
        when(xrayAddon.isXrayEnabled()).thenReturn(true);
        when(xrayAddon.isXrayAliveByHeartbeat()).thenReturn(true);
        when(xrayAddon.isXrayVersionValidByHeartbeat()).thenReturn(false);

        String globalMessages = msgProvider.getGlobalMessages();
        List<FooterMessage> footerMessages = JsonUtils.getInstance()
                .readValueRef(globalMessages, new TypeReference<>() {
                });
        assertNotNull(footerMessages);
        assertFalse(footerMessages.isEmpty());
        assertThat(footerMessages.get(0).getMessage()).contains("is incompatible with this");

        // Change state of Xray and validate message is still the same (since did not invalidate cache)
        when(xrayAddon.isXrayVersionValidByHeartbeat()).thenReturn(true);
        globalMessages = msgProvider.getGlobalMessages();
        footerMessages = JsonUtils.getInstance()
                .readValueRef(globalMessages, new TypeReference<>() {
                });
        assertNotNull(footerMessages);
        assertFalse(footerMessages.isEmpty());

        // Invalidate and assert we now don't get any errors
        msgProvider.invalidateMessagesCache();
        globalMessages = msgProvider.getGlobalMessages();
        footerMessages = JsonUtils.getInstance()
                .readValueRef(globalMessages, new TypeReference<>() {
                });
        assertNotNull(footerMessages);
        assertTrue(footerMessages.isEmpty());
    }

    //FooterMessageVisibility.all should be shown to anon
    public void testAnonymousCanSeeAllVisibility() {
        reset(authService);
        when(authService.isAdmin()).thenReturn(false);
        when(authService.isAnonymous()).thenReturn(true);

        FooterMessage anonVisibleMsg = createWarning("Anonymous can see me", null, FooterMessageVisibility.all);
        when(addonsManager.getLicenseFooterMessage()).thenReturn(anonVisibleMsg);

        assertThat(msgProvider.getGlobalMessages()).isEqualTo(toJson(anonVisibleMsg));
    }

    //FooterMessageVisibility.user should be shown to non admin, non anon
    public void testUserCanSeeUserVisibility() {
        reset(authService);
        when(authService.isAdmin()).thenReturn(false);
        when(authService.isAnonymous()).thenReturn(false);

        FooterMessage anonVisibleMsg = createWarning("user can see me", List.of(), FooterMessageVisibility.all);
        when(addonsManager.getLicenseFooterMessage()).thenReturn(anonVisibleMsg);

        assertThat(msgProvider.getGlobalMessages()).isEqualTo(toJson(anonVisibleMsg));
    }

    public void testMessageWithoutRange() throws IOException {
        FooterMessage msg = FooterMessage.builder()
                .message("meow")
                .type("info")
                .visibility("user")
                .build();
        setCustomMessage(msg);

        assertThat(msgProvider.getGlobalMessages()).isEqualTo(toJson(msg));
    }

    public void testCustomMessageCache() throws IOException {
        assertNoMessages();
        FooterMessage msg = FooterMessage.builder()
                .message("meow")
                .type("info")
                .visibility("user")
                .build();
        setCustomMessage(msg);
        // message will not be updated for an hour
        assertNoMessages();
    }

    public void testMessageInRange() throws IOException {
        FooterMessage msg = FooterMessage.builder()
                .message("how")
                .type("info")
                .visibility("user")
                .showFrom(System.currentTimeMillis() - 1L)
                .showUntil(System.currentTimeMillis() + 1000000L)
                .build();
        setCustomMessage(msg);

        assertThat(msgProvider.getGlobalMessages()).isEqualTo(toJson(msg));
    }

    //A message with a range that has already passed (should not show)
    public void testMessageWithPastRange() throws IOException {
        FooterMessage msg = FooterMessage.builder()
                .message("This is a test of the emergency broadcast system, please stand by.")
                .type("info")
                .visibility("user")
                .showFrom(System.currentTimeMillis() - 10000000L)
                .showUntil(System.currentTimeMillis() - 1000L)
                .build();
        setCustomMessage(msg);

        assertNoMessages();
    }

    //A message with a range that was not yet reached (should not show)
    public void testMessageWithFutureRange() throws IOException {
        FooterMessage msg = FooterMessage.builder()
                .message("I feel good today, Silent Bob. We're gonna make some money...")
                .type("info")
                .visibility("user")
                .showFrom(System.currentTimeMillis() + 10000000L)
                .showUntil(System.currentTimeMillis() + 10000000000L)
                .build();
        setCustomMessage(msg);

        assertNoMessages();
    }

    public void testDismissCode() throws IOException {
        FooterMessage msg = FooterMessage.builder()
                .message("My dismiss code bring all the bots to the yard")
                .type("info")
                .visibility("user")
                .dismissible(true)
                .build();

        setCustomMessage(msg);
        assertThat(msgProvider.getGlobalMessages())
                .contains("\"dismissCode\":\"331d7e580da8973a49b2c4565eb7323f\"");
    }

    private void setCustomMessage(FooterMessage msg) throws IOException {
        Files.write(toJson(msg).getBytes(), tempFile);
        homeStub.setProperty(ConstantValues.uiCustomFooterMessageFile, tempFile.getAbsolutePath());
    }

    private void assertNoMessages() {
        try {
            JSONAssert.assertEquals(msgProvider.getGlobalMessages(), "[]", true);
        } catch (JSONException e) {
            throw new IllegalArgumentException(e);
        }
    }

    private String toJson(FooterMessage msg) {
        return JsonUtils.getInstance().valueToString(ImmutableList.of(msg));
    }
}